/*! \file HuffwordLib.h

  \brief Interface of C Library processing.
  
  \author  Paolo Ferragina, Dipartimento di Informatica, Pisa (Italy).

  \date November 2002.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef WIN32
	#include <sys/mman.h>
	#include <unistd.h>
#else
	#include <Windows.h>
#endif
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>
#include <math.h>


/* -------------------------------------------------------- */
/* ---------------  Pull of data structures --------------- */
/* -------------------------------------------------------- */



/*! Element of the catenating lists in the hash table. the field
  "codeword" contains the byte-aligned and tagged codeword of the
  token stored in "string", aligned rightward. The codeword length in
  bytes is kept into the field "cw_len". */ 
typedef struct Hash_node {
  char	*str;          //!< token
  int len_str;         //!< length of the token (to manage also NULL)
  int count_occ;       //!< number of occurrences of the token
  int codeword;        //!< codeword for this token byte-aligned and tagged
  int cw_len;          //!< length of the codeword in bytes
  struct Hash_node *next;    //!< pointer to the next item in the list
} Hash_node;

typedef Hash_node **Hash_nodeptr_array; //!< array of pointers to Hash_nodes

/*! Data structure for an hash table managed by catenating lists and
  MTF. The field "size" indicates the table size, whereas the field
  "card" indicates the number of objects stored into the catenating
  lists. */ 
typedef struct {
  int size;             //!< hash table size
  int card;             //!< number of stored items
  Hash_nodeptr_array table;   //!< pointer to the hash table
} HHash_table;



/*! Data structure containing the dictionary and its associated
  information. The tokens are formed in three distinct ways. Either as
  sequences of alphanumeric chars, or as a sequence of separators (but
  not '\n'), or as a singleton '\n'. This choice is dictated by the
  way the dictionary string is manipulated and stored, it assumes that
  each token is terminated by '\n'.  */ 
typedef struct {
  char *content;    //!< string obtained concatenating the tokens separated by NEWLINE
  int length;       //!< length of the dictionary string
  int num_tokens;   //!< number of the dictionary tokens
  int *start_pos;   //!< array of starting positions in "string" of the various tokens
} Dictionary;



/*! Data structure containing the canonical huffman information.  */ 
typedef struct {
  int firstcw[5];   //!< first codeword of each level (from 1 to 4)
  int offsetcw[5];  //!< dictionary position of the first token of levels 1...4, it is > num_tokens if absent
  int max_cwlen;    //!< maximum codeword length in bytes
  int num_tokens;   //!< number of tokens in the Dictionary data structure
} Canonical;



/*! Data structure containing the pointers to some positions in the compressed text. */ 
typedef struct {
  int number;             //!< number of defined jumpers
  int *text_offsets;      //!< array of starting positions in the plain text
  int *ctext_offsets;     //!< array of starting positions in the compressed text (aligned to cw starts)
} Jumpers;


/*! All data structures in a nutshell. */ 
typedef struct {
  HHash_table hashtable;           //!< hash table
  Dictionary dictionary;          //!< dictionary infos
  Canonical canoinfo;   //!< canonical huffman infos
  Jumpers jumpers;       //!< jumpers into compressed text
} Console;



/* ------------------------------------------------------ */
/* -----------  Text parsing and compression ------------ */
/* ------------------------------------------------------ */


void HParse_text(char *text, int text_len, HHash_table *ht);

void Huffw_compress(char *text, int tlen, int jump_value, char **cstring, int *clen, int Verbose);
void Huffw_decompress(char *cstring, int clen, char **text, int *tlen, int Verbose);
void Huffw_PrintInfo(Console *console, int ctext_len, int text_len, int cstring_len, 
		     int rule, int Verbose);

void Huffw_spaceless_compress(char *text, int tlen, char **cstring, int *clen, int Verbose);
void Huffw_spaceless_decompress(char *cstring, int clen, char **text, int *tlen, int Verbose);


void HCompress_getBody(char *text, int text_len, HHash_table *ht, char **ctext, int *ctext_len);
void HCompress_getBodyAndConsole(char *text, int tlen, int jump_value, char **body, int *blen, 
				Console *console
#ifdef WIN32
				, Hash_node ***to_free
#endif
				);
void HCompress_getString(char **ctext, int *ctext_len, int text_len,char *body, 
			     int blen, Console *console);
void HCompress_contractSpaces(char *body, int blen, char **body_spaceless, int *blen_spaceless, 
				  Console *console);


void HDecompress_getBodyAndConsole(char *ctext, int ctext_len, int *text_len, char **body, 
				       int *blen, Console *console);
void HDecompress_expandSpaces(char *body_spaceless, int blen_spaceless, char **body, 
				  int *blen, Console *console);
void HDecompress_nextBlock_bytes(char *ctext, int num_bytes, char **text, int *text_len, 
				     Console *console);
void HDecompress_nextBlock_bytes_spaceless(char *ctext, int num_bytes, char **text, 
					       int *text_len, Console *console);
void HDecompress_previousBlock_bytes(char *ctext, int num_bytes, char **text, int *text_len, 
					 Console *console);
void HDecompress_previousBlock_bytes_spaceless(char *ctext, int num_bytes, char **text, 
						   int *text_len, Console *console);
void HDecompress_nextBlock_tokens(char *ctext, int byte_left, int num_obj, char **text, 
				      int *text_len, int *decoded, Console *console);
void HDecompress_nextBlock_tokens_spaceless(char *ctext, int bytes_left, int num_obj, 
						char **text, int *text_len, int *decoded,
						Console *console);
void HDecompress_previousBlock_tokens(char *ctext, int byte_left, int num_obj, char **text, 
					  int *text_len, int *decoded, Console *console);
void HDecompress_previousBlock_tokens_spaceless(char *ctext, int bytes_left, int num_obj, 
						    char **text, int *text_len, int *decoded,
						    Console *console);




/* ---------------------------------------------------- */
/* -------------------  Hash Table    ----------------- */
/* ---------------------------------------------------- */


void HHashtable_init(HHash_table *ht, int estimated);
int HHashtable_func(char *s, int len, HHash_table *ht);
Hash_node *HHashtable_search(char *str, int len, HHash_table *ht);
int HHashtable_insert(char *str, int str_len, HHash_table *ht);   
void HHashtable_clear(HHash_table *ht);
void HHashtable_fromdict(Dictionary *dict, Canonical *cano, HHash_table *ht);
void HHashtable_print(HHash_table *ht);


/* ------------------------------------------------------------ */
/* ---------------------  Huffman Tree   ---------------------- */
/* ------------------------------------------------------------ */

void Hufftree_createLeaves(Hash_nodeptr_array tree, int num_leaves, HHash_table *ht);
int Hufftree_fromLeaves(Hash_nodeptr_array work_area, int *tot_nodes, int HT_leaves);
int Hufftree_computeCwLen(Hash_nodeptr_array tree, int root, int tree_size);
void Hufftree_build(Hash_nodeptr_array *tree_array_ptr, HHash_table *ht
#ifdef WIN32
					, Hash_node ***to_free
#endif
					);
void Hufftree_print(Canonical *cano, Hash_nodeptr_array tree_array, int Verbose);



/* ------------------------------------------------------ */
/* -----------------  Codeword operations --------------- */
/* ------------------------------------------------------ */

void HCodeword_PlainFromTagged(int taggedcw, int taggedcwlen,int *plaincw, int *plaincwlen);
void HCodeword_PlainFromTokenrank(int token_rank,Canonical *cano,int *plaincw,int *plaincwlen); 
void HCodeword_TaggedFromPlain(int plaincw, int plaincwlen,int *taggedcw, int *taggedcwlen); 
void HCodeword_TaggedFromTokenrank(int token_rank,Canonical *cano,int *taggedcw, int *taggedcwlen); 
void HCodeword_TaggedFromToken(char *token, int token_len, HHash_table *ht,int *taggedcw, int *taggedcwlen); 
void HCodeword_TaggedGetNext(char *s, int num_byte_left, int *taggedcw, int *lencw);
int HCodeword_TaggedGetNextLength(char *s, int num_byte_left);
void HCodeword_TaggedGetPrevious(char *s, int num_byte_left, int *taggedcw, int *lencw);
void HCodeword_tostring(int taggedcw, int taggedcwlen, char **s); 

void HToken_RankFromPlainCw(int plaincw, int plaincwlen, Canonical *cano,int *token_rank); 
void HToken_RankFromTaggedCw(int taggedcw, int taggedcwlen, Canonical *cano,int *token_rank); 
void HToken_fromTaggedCw(int taggedcw, int taggedcwlen,Canonical *cano, Dictionary *dict,
			     char **token, int *token_len);
int HToken_decompressNext(char *s, int num_byte_left, char **token, int *lentoken, int *lencw, 
			 Console *console);
int HToken_decompressPrevious(char *s, int num_byte_left, char **token, int *lentoken, int *lencw, 
			     Console *console);

void HToken_getLengthNext(char *s, int num_char_left, int *len);


/* ------------------------------------------------------ */
/* ---------------  Dictionary operations --------------- */
/* ------------------------------------------------------ */

void HDictionary_fromstring(char *s, int slen, int stokens, Dictionary *dict);
void HDictionary_fromtree(Hash_nodeptr_array tree, int num_tokens, Dictionary *dict);
void HDictionary_print(Dictionary *dict, HHash_table *ht, int Verbose);

/* -------------------------------------------------------- */
/* ----------------  Jumpers data structure --------------- */
/* -------------------------------------------------------- */

void HJumpers_fromds(char *text, int tlen, HHash_table *ht,int jump_value, Jumpers *jumpers);
void HJumpers_tostring(char *s, Jumpers *jumpers);
void HJumpers_fromstring(char *s, Jumpers *jumpers);
void HGet_charpos_from_bytepos(char *ctext, int ctext_len, int bytepos, int *textpos, Console *console);
void HGet_bytepos_from_charpos(char *ctext, int ctext_len, int textpos, int *bytepos, Console *console);

/* ---------------------------------------------------------- */
/* ----------------  Canonical data structure --------------- */
/* ---------------------------------------------------------- */


void HCanonical_fromtree(Hash_nodeptr_array tree_array, Canonical *cano, HHash_table *ht);
void HCanonical_fromstring(char *s, Canonical *cano);
void HCanonical_tostring(char *s, Canonical *cano);





/* ---------------------------------------------------- */
/* -----------------  Basic routines    --------------- */
/* ---------------------------------------------------- */


/*! Stores the integer "i" into the first 4 bytes of "s" */
void HInt_tostring(char *s, int i);


/*! Reads the first 4 bytes of "s" as an integer. */
int HInt_fromstring(char *s); 

/*! Auxiliary print function */
void HPrint_string(char *s, int l);

/* Used in the construction of the Huffman tree */
int HSort_for_freq(const void *va, const void *vb);
int HSort_for_cwlen(const void *va, const void *vb);


/* ---------------- Code for freeing allocated memory ---------------- */
#ifdef WIN32
void FreeTheConsole(Console *console, int freeDictionaryContent);
void FreeDictionary(Dictionary *dictionary, int freeDictionaryContent);
void FreeHashtable(HHash_table *hashtable);
void FreeJumpers(Jumpers *jumpers);
#endif